/****************************************************************************
*   PROJECT: Balloon 3D Graphics Subsystem for Squeak
*   FILE:    b3dDraw.c
*   CONTENT: Pixel drawing functions for the B3D rasterizer
*
*   AUTHOR:  Andreas Raab (ar)
*   ADDRESS: Walt Disney Imagineering, Glendale, CA
*   EMAIL:   andreasr@wdi.disney.com
*   RCSID:   $Id$
*
*   NOTES:  LOTS of stuff missing here...
*
*
*****************************************************************************/
#include "b3d.h"

#define rasterPosX rasterPos[0]
#define rasterPosY rasterPos[1]

#define redValue   color[RED_INDEX]
#define greenValue color[GREEN_INDEX]
#define blueValue  color[BLUE_INDEX]
#define alphaValue color[ALPHA_INDEX]

#ifdef DEBUG_ATTR
double attrValueAt(B3DPrimitiveFace *face, B3DPrimitiveAttribute *attr, double xValue, double yValue)
{
	return 
		(attr->value +
			((xValue - face->v0->rasterPosX) * attr->dvdx) +
			((yValue - face->v0->rasterPosY) * attr->dvdy));
}
#else
#define attrValueAt(face,attr,xValue,yValue) \
		((attr)->value + \
			(((double)(xValue) - (face)->v0->rasterPosX) * (attr)->dvdx) + \
			(((double)(yValue) - (face)->v0->rasterPosY) * (attr)->dvdy))
#endif


#define SETUP_RGB \
	rValue = (int)(attrValueAt(face, attr, floatX, floatY) * B3D_FloatToFixed); \
	deltaR = (int) (attr->dvdx * B3D_FloatToFixed);  \
	if(rValue < B3D_FixedHalf) rValue = B3D_FixedHalf; \
	if(rValue > (255 << B3D_IntToFixedShift) + B3D_FixedHalf) rValue = (255 << B3D_IntToFixedShift) + B3D_FixedHalf; \
	attr = attr->next; \
	gValue = (int)(attrValueAt(face, attr, floatX, floatY) * B3D_FloatToFixed);\
	deltaG = (int) (attr->dvdx * B3D_FloatToFixed); \
	if(gValue < B3D_FixedHalf) gValue = B3D_FixedHalf; \
	if(gValue > (255 << B3D_IntToFixedShift) + B3D_FixedHalf) gValue = (255 << B3D_IntToFixedShift) + B3D_FixedHalf; \
	attr = attr->next; \
	bValue = (int)(attrValueAt(face, attr, floatX, floatY) * B3D_FloatToFixed); \
	deltaB = (int) (attr->dvdx * B3D_FloatToFixed); \
	if(bValue < B3D_FixedHalf) bValue = B3D_FixedHalf; \
	if(bValue > (255 << B3D_IntToFixedShift) + B3D_FixedHalf) bValue = (255 << B3D_IntToFixedShift) + B3D_FixedHalf; \
	attr = attr->next;

#define SETUP_ALPHA \
	aValue = (int)(attrValueAt(face, attr, floatX, floatY) * B3D_FloatToFixed); \
	deltaA = (int) (attr->dvdx * B3D_FloatToFixed);  \
	if(aValue < B3D_FixedHalf) aValue = B3D_FixedHalf; \
	if(aValue > (255 << B3D_IntToFixedShift) + B3D_FixedHalf) aValue = (255 << B3D_IntToFixedShift) + B3D_FixedHalf; \
	attr = attr->next;

#define SETUP_STW \
	wValue = attrValueAt(face, attr, floatX, floatY); \
	wDelta = attr->dvdx; \
	attr = attr->next; \
	sValue = attrValueAt(face, attr, floatX, floatY); \
	sDelta = attr->dvdx; \
	attr = attr->next; \
	tValue = attrValueAt(face, attr, floatX, floatY); \
	tDelta = attr->dvdx; \
	attr = attr->next;

#define STEP_RGB \
	rValue += deltaR;\
	if(rValue < B3D_FixedHalf) rValue = B3D_FixedHalf;\
	if(rValue > (255 << B3D_IntToFixedShift) + B3D_FixedHalf) rValue = (255 << B3D_IntToFixedShift) + B3D_FixedHalf;\
	gValue += deltaG;\
	if(gValue < B3D_FixedHalf) gValue = B3D_FixedHalf;\
	if(gValue > (255 << B3D_IntToFixedShift) + B3D_FixedHalf) gValue = (255 << B3D_IntToFixedShift) + B3D_FixedHalf;\
	bValue += deltaB;\
	if(bValue < B3D_FixedHalf) bValue = B3D_FixedHalf;\
	if(bValue > (255 << B3D_IntToFixedShift) + B3D_FixedHalf) bValue = (255 << B3D_IntToFixedShift) + B3D_FixedHalf;

#define STEP_ALPHA \
	aValue += deltaA;\
	if(aValue < B3D_FixedHalf) aValue = B3D_FixedHalf;\
	if(aValue > (255 << B3D_IntToFixedShift) + B3D_FixedHalf) aValue = (255 << B3D_IntToFixedShift) + B3D_FixedHalf;


#define STEP_STW \
	sValue += sDelta;\
	tValue += tDelta;\
	wValue += wDelta;


#define LOAD_TEXTURE_32 \
{\
	double fs, ft, oneOverW;\
	int ffs, fft;\
	struct b3dPixelColor *tex00, *tex01, *tex10, *tex11;\
	int sf, si, tf, ti;\
\
	if(wValue) oneOverW = 1.0 / wValue;\
	fs = sValue * oneOverW * (texture->width);\
	ft = tValue * oneOverW * (texture->height);\
	ffs = (int)fs % (texture->width);\
	fft = (int)ft % (texture->height);\
	tex00 = (struct b3dPixelColor *)texture->data + (fft * texture->width) + ffs;\
	tex01 = tex00 + 1;\
	tex10 = tex00 + texture->width;\
	tex11 = tex10 + 1;\
	if(ffs+1 == texture->width) {\
		tex01 -= texture->width;\
		tex11 -= texture->width;\
	}\
	if(fft+1 == texture->height) {\
		int tsize = texture->height * texture->width;\
		tex10 -= tsize;\
		tex11 -= tsize;\
	}\
	sf = (int) (fs * 16.0) & 15; si = 15 - sf;\
	tf = (int) (ft * 16.0) & 15; ti = 15 - tf;\
	tr = (ti * (si * tex00->redValue + sf * tex01->redValue) +\
			tf * (si * tex10->redValue + sf * tex11->redValue)) >> 8;\
	tg = (ti * (si * tex00->greenValue + sf * tex01->greenValue) +\
			tf * (si * tex10->greenValue + sf * tex11->greenValue)) >> 8;\
	tb = (ti * (si * tex00->blueValue + sf * tex01->blueValue) +\
			tf * (si * tex10->blueValue + sf * tex11->blueValue)) >> 8;\
	ta = (ti * (si * tex00->alphaValue + sf * tex01->alphaValue) +\
			tf * (si * tex10->alphaValue + sf * tex11->alphaValue)) >> 8;\
}


void b3dNoDraw     (int leftX, int rightX, int yValue, B3DPrimitiveFace *face);
void b3dDrawRGB    (int leftX, int rightX, int yValue, B3DPrimitiveFace *face);
void b3dDrawRGBA   (int leftX, int rightX, int yValue, B3DPrimitiveFace *face);
void b3dDrawSTW    (int leftX, int rightX, int yValue, B3DPrimitiveFace *face);
void b3dDrawSTWA   (int leftX, int rightX, int yValue, B3DPrimitiveFace *face);
void b3dDrawSTWRGB (int leftX, int rightX, int yValue, B3DPrimitiveFace *face);
void b3dDrawSTWARGB(int leftX, int rightX, int yValue, B3DPrimitiveFace *face);

b3dPixelDrawer B3D_FILL_FUNCTIONS[B3D_MAX_ATTRIBUTES] = {
	b3dNoDraw,      /* No attributes */
	b3dDrawRGB,     /* B3D_FACE_RGB */
	b3dNoDraw,      /* B3D_FACE_ALPHA -- IGNORED!!! */
	b3dDrawRGBA,    /* B3D_FACE_RGB | B3D_FACE_ALPHA */
	b3dDrawSTW,     /* B3D_FACE_STW */
	b3dDrawSTWRGB,  /* B3D_FACE_STW | B3D_FACE_RGB */
	b3dDrawSTWA,    /* B3D_FACE_STW | B3D_FACE_ALPHA */
	b3dDrawSTWARGB  /* B3D_FACE_STW | B3D_FACE_RGB | B3D_FACE_ALPHA */
};

void b3dNoDraw(int leftX, int rightX, int yValue, B3DPrimitiveFace *face)
{
	if(b3dDebug)
		b3dAbort("b3dNoDraw called!");
}

void b3dDrawRGB(int leftX, int rightX, int yValue, B3DPrimitiveFace *face)
{
	struct b3dPixelColor { B3DPrimitiveColor color; } pv, *bits;
	int rValue, gValue, bValue;
	int deltaR, deltaG, deltaB;


	{
		B3DPrimitiveAttribute *attr = face->attributes;
		/* Ughh ... I'm having a sampling problem somewhere.
		   In theory, the faces should be sampled *exactly* at integer
		   values (the necessary offset should be done before) so that
		   we always sample inside the triangle. For some reason that
		   doesn't quite work yet and that's why here is the strange
		   0.5 offset and the awful lot of tests. At some time I'll
		   review this but for now I have more important things to do.
		*/
		double floatX = leftX;
		double floatY = yValue+0.5;

		if(b3dDebug)
			if(!attr) b3dAbort("face has no RGB attributes");

		SETUP_RGB;
	}

	bits = (struct b3dPixelColor *) currentState->spanBuffer;
	pv.alphaValue = 255;

	while(leftX <= rightX) {
		pv.redValue   = (unsigned char) (rValue >> B3D_FixedToIntShift);
		pv.greenValue = (unsigned char) (gValue >> B3D_FixedToIntShift);
		pv.blueValue  = (unsigned char) (bValue >> B3D_FixedToIntShift);
		bits[leftX++] = pv;
		STEP_RGB;
	}
}

void b3dDrawRGBA(int leftX, int rightX, int yValue, B3DPrimitiveFace *face)
{
	b3dDrawRGB(leftX, rightX, yValue, face);
	/* not yet implemented */
}

void b3dDrawSTW(int leftX, int rightX, int yValue, B3DPrimitiveFace *face)
{
	struct b3dPixelColor { B3DPrimitiveColor color; } pv, *bits;
	double sValue, tValue, wValue, sDelta, tDelta, wDelta;
	int tr, tg, tb, ta;
	B3DTexture *texture = face->texture;

	if(!texture) return; /* Nothing to do if no texture at all is given */
	if(texture->depth < 16 && (texture->cmSize < (1 << texture->depth)))
		return; /* Colormap not installed */

	{
		B3DPrimitiveAttribute *attr = face->attributes;
		/* See above */
		double floatX = leftX;
		double floatY = yValue+0.5;

		if(b3dDebug)
			if(!attr) b3dAbort("face has no RGB attributes");

		SETUP_STW;
	}

	bits = (struct b3dPixelColor *) currentState->spanBuffer;
	pv.redValue = 0;
	pv.greenValue = 0;
	pv.blueValue = 0;
	pv.alphaValue = 255;

	while(leftX <= rightX) {
		if(texture->depth < 16) {
			/* Colormap lookup */
		} else {
			/* Direct color */
			if(texture->depth == 32) {
				LOAD_TEXTURE_32;
			} else {
				/* 16 bit texture */
			}
		}
		pv.redValue = (unsigned char) tr;
		pv.greenValue = (unsigned char) tg;
		pv.blueValue = (unsigned char) tb;
		bits[leftX++] = pv;
		STEP_STW;
	}
}

void b3dDrawSTWA(int leftX, int rightX, int yValue, B3DPrimitiveFace *face)
{
	B3DPrimitiveAttribute *attr = face->attributes;
	/* Skip alpha -- this is not yet implemented */
	face->attributes = face->attributes->next;
	b3dDrawSTW(leftX, rightX, yValue, face);
	face->attributes = attr;
}

void b3dDrawSTWRGB(int leftX, int rightX, int yValue, B3DPrimitiveFace *face)
{
	struct b3dPixelColor { B3DPrimitiveColor color; } pv, *bits;
	double sValue, tValue, wValue, sDelta, tDelta, wDelta;
	int rValue, gValue, bValue;
	int deltaR, deltaG, deltaB;
	int tr, tg, tb, ta;
	B3DTexture *texture = face->texture;

	if(!texture) return; /* Nothing to do if no texture at all is given */
	if(texture->depth < 16 && (texture->cmSize < (1 << texture->depth)))
		return; /* Colormap not installed */

	{
		B3DPrimitiveAttribute *attr = face->attributes;
		/* See above */
		double floatX = leftX;
		double floatY = yValue+0.5;

		if(b3dDebug)
			if(!attr) b3dAbort("face has no RGB attributes");

		SETUP_RGB;
		SETUP_STW;
	}

	bits = (struct b3dPixelColor *) currentState->spanBuffer;
	pv.redValue = 0;
	pv.greenValue = 0;
	pv.blueValue = 0;
	pv.alphaValue = 255;

	while(leftX <= rightX) {
		if(texture->depth < 16) {
			/* Colormap lookup */
		} else {
			/* Direct color */
			if(texture->depth == 32) {
				LOAD_TEXTURE_32;
			} else {
				/* 16 bit texture */
			}
		}
		pv.redValue   = (unsigned char) ((tr * rValue) >> (B3D_FixedToIntShift + 8));
		pv.greenValue = (unsigned char) ((tg * gValue) >> (B3D_FixedToIntShift + 8));
		pv.blueValue  = (unsigned char) ((tb * bValue) >> (B3D_FixedToIntShift + 8));
		bits[leftX++] = pv;
		STEP_RGB;
		STEP_STW;
	}
}

void b3dDrawSTWARGB(int leftX, int rightX, int yValue, B3DPrimitiveFace *face)
{
	B3DTexture *texture = face->texture;	
	if(!texture) {
		/* If no texture simply draw RGB */
		b3dDrawRGBA(leftX, rightX, yValue, face);
	}
	{
		B3DPrimitiveAttribute *attr = face->attributes;
		/* Skip RGB */
		face->attributes = face->attributes->next->next->next;
		b3dDrawSTWA(leftX, rightX, yValue, face);
		face->attributes = attr;
	}
}
